@@ -24,16 +24,23 @@ extends qw(
   Lemonldap::NG::Common::Conf::RESTServer
 );
 
-our $VERSION = '2.0.12';
+our $VERSION = '2.19.0';
 
 #############################
 # I. INITIALIZATION METHODS #
 #############################
 
 use constant defaultRoute => 'manager.html';
+use constant icon         => 'cog';
+
+has defaultNewKeySize => ( is => 'rw', default => 2048, );
 
 sub init {
     my ( $self, $conf ) = @_;
+
+    $self->defaultNewKeySize( $conf->{defaultNewKeySize} )
+      if $conf->{defaultNewKeySize};
+
     $self->ua( Lemonldap::NG::Common::UserAgent->new($conf) );
 
     # HTML template
@@ -62,6 +69,7 @@ sub init {
         confs => {
             newRSAKey      => 'newRSAKey',
             newCertificate => 'newCertificate',
+            newEcKeys      => 'newEcKeys',
             sendTestMail   => 'sendTestMail',
             raw            => 'newRawConf',
             '*'            => 'newConf'
@@ -99,6 +107,7 @@ sub newCertificate {
     my $keys = {
         'private' => $private,
         'public'  => $cert,
+        'hash'    => md5_base64($cert),
     };
     return $self->sendJSONresponse( $req, $keys );
 }
@@ -117,7 +126,7 @@ sub newRSAKey {
     my ( $self, $req, @others ) = @_;
     return $self->sendError( $req, 'There is no subkey for "newRSAKey"', 400 )
       if (@others);
-    my $rsa = Crypt::OpenSSL::RSA->generate_key(2048);
+    my $rsa = Crypt::OpenSSL::RSA->generate_key( $self->defaultNewKeySize );
 
     my $query = $req->jsonBodyToObj;
     my $keys  = {
@@ -150,6 +159,30 @@ sub newRSAKey {
     return $self->sendJSONresponse( $req, $keys );
 }
 
+# 35 - New EC key pair on demand
+#      --------------------------
+
+##@method public PSGI-JSON-response newEcKeys($req)
+# Return a hashref containing private and public keys
+#
+#@param $req Lemonldap::NG::Common::PSGI::Request object
+#@return PSGI JSON response
+sub newEcKeys {
+    my ( $self, $req, @others ) = @_;
+    require Crypt::PK::ECC;
+    my $ec_key = Crypt::PK::ECC->new();
+    $ec_key->generate_key('secp256r1');
+
+    my $pubKey = $ec_key->export_key_pem('public');
+    my $privKey = $ec_key->export_key_pem('private');
+    my $keys = {
+        private => $privKey,
+        public => $pubKey,
+        hash => md5_base64($pubKey),
+    };
+    return $self->sendJSONresponse( $req, $keys );
+}
+
 # This function does the dirty X509 work,
 # mostly copied from IO::Socket::SSL::Utils
 # and adapter to work on old platforms (CentOS7)
@@ -162,7 +195,7 @@ sub _generateX509 {
     # Generate 2048 bits RSA key
     my $key = Net::SSLeay::EVP_PKEY_new();
     Net::SSLeay::EVP_PKEY_assign_RSA( $key,
-        Net::SSLeay::RSA_generate_key( 2048, 0x10001 ) );
+        Net::SSLeay::RSA_generate_key( $self->defaultNewKeySize, 0x10001 ) );
 
     my $cert = Net::SSLeay::X509_new();
 
@@ -212,7 +245,10 @@ sub _generateX509 {
     my $strCert = Net::SSLeay::PEM_get_string_X509($cert);
     my $strPrivate;
     if ($password) {
-        $strPrivate = Net::SSLeay::PEM_get_string_PrivateKey( $key, $password );
+        my $alg = Net::SSLeay::EVP_get_cipherbyname("AES-256-CBC")
+          || Net::SSLeay::EVP_get_cipherbyname("DES-EDE3-CBC");
+        $strPrivate =
+          Net::SSLeay::PEM_get_string_PrivateKey( $key, $password, $alg );
     }
     else {
         $strPrivate = Net::SSLeay::PEM_get_string_PrivateKey($key);
@@ -393,8 +429,11 @@ sub newConf {
     $res->{details}->{'__errors__'} = $parser->{errors}
       if ( @{ $parser->{errors} } );
     unless ( @{ $parser->{errors} } ) {
-        $res->{details}->{'__needConfirmation__'} = $parser->{needConfirmation}
-          if ( @{ $parser->{needConfirmation} } && !$req->params('force') );
+        if ( @{ $parser->{needConfirmation} } && !$req->params('force') ) {
+            $res->{needConfirm} = 1;
+            $res->{details}->{'__needConfirmation__'} =
+              $parser->{needConfirmation};
+        }
         $res->{message} = $parser->{message};
         foreach my $t (qw(warnings changes)) {
             $res->{details}->{ '__' . $t . '__' } = $parser->$t
@@ -408,12 +447,20 @@ sub newConf {
             $args{cfgDate}        = $req->params('cfgDate');
             $args{currentCfgDate} = $currentCfgDate;
         }
-        my $s = CONFIG_WAS_CHANGED;
+        my $s = UNKNOWN_ERROR;
         $s = $self->confAcc->saveConf( $parser->newConf, %args )
           unless ( @{ $parser->{needConfirmation} } && !$args{force} );
         if ( $s > 0 ) {
-            $self->userLogger->notice(
-                'User ' . $self->p->userId($req) . " has stored conf $s" );
+            $self->auditLog(
+                $req,
+                message => (
+                    'User ' . $self->p->userId($req) . " has stored conf $s"
+                ),
+                code   => "CONF_STORED",
+                user   => $self->p->userId($req),
+                cfgNum => $s,
+            );
+
             $res->{result} = 1;
             $res->{cfgNum} = $s;
             if ( my $status = $self->applyConf( $parser->newConf ) ) {
@@ -423,14 +470,22 @@ sub newConf {
             }
         }
         else {
-            $self->userLogger->notice(
-                'Saving attempt rejected, asking for confirmation to '
-                  . $self->p->userId($req) );
+            $self->auditLog(
+                $req,
+                message => (
+                    'Saving attempt rejected, asking for confirmation to '
+                      . $self->p->userId($req)
+                ),
+                code => "CONF_REJECTED",
+                user => $self->p->userId($req),
+            );
+
             $res->{result} = 0;
             if ( $s == CONFIG_WAS_CHANGED ) {
                 $res->{needConfirm} = 1;
-                $res->{message} .= '__needConfirmation__'
-                  unless @{ $parser->{needConfirmation} };
+                $res->{details}->{'__needConfirmation__'} ||= [];
+                push @{ $res->{details}->{'__needConfirmation__'} },
+                  { message => '__newCfgAvailableWarning__' };
             }
             else {
                 $res->{message} = $Lemonldap::NG::Common::Conf::msg;
@@ -457,28 +512,36 @@ sub newRawConf {
     }
 
     my $res = {};
-    if ( $self->p->{demoMode} ) {
-        $res->{message} = '__demoModeOn__';
+
+    # When uploading a new conf, always force it since cfgNum has a few
+    # chances to be equal to last config cfgNum
+    my $s = $self->confAcc->saveConf( $new, force => 1 );
+    if ( $s > 0 ) {
+        $self->auditLog(
+            $req,
+            message => (
+                'User ' . $self->p->userId($req) . " has stored (raw) conf $s"
+            ),
+            code   => "CONF_STORED_RAW",
+            user   => $self->p->userId($req),
+            cfgNum => $s,
+        );
+        $res->{result} = 1;
+        $res->{cfgNum} = $s;
     }
     else {
-        # When uploading a new conf, always force it since cfgNum has a few
-        # chances to be equal to last config cfgNum
-        my $s = $self->confAcc->saveConf( $new, force => 1 );
-        if ( $s > 0 ) {
-            $self->userLogger->notice( 'User '
-                  . $self->p->userId($req)
-                  . " has stored (raw) conf $s" );
-            $res->{result} = 1;
-            $res->{cfgNum} = $s;
-        }
-        else {
-            $self->userLogger->notice(
+        $self->auditLog(
+            $req,
+            message => (
                 'Raw saving attempt rejected, asking for confirmation to '
-                  . $self->p->userId($req) );
-            $res->{result}      = 0;
-            $res->{needConfirm} = 1 if ( $s == CONFIG_WAS_CHANGED );
-            $res->{message} .= '__needConfirmation__';
-        }
+                  . $self->p->userId($req)
+            ),
+            code => "CONF_REJECTED_RAW",
+            user => $self->p->userId($req),
+        );
+        $res->{result}      = 0;
+        $res->{needConfirm} = 1 if ( $s == CONFIG_WAS_CHANGED );
+        $res->{message} .= '__needConfirmation__';
     }
     return $self->sendJSONresponse( $req, $res );
 }
