@@ -2,7 +2,7 @@
 # Display functions for LemonLDAP::NG Portal
 package Lemonldap::NG::Portal::Main::Display;
 
-our $VERSION = '2.16.1';
+our $VERSION = '2.19.0';
 
 package Lemonldap::NG::Portal::Main;
 use strict;
@@ -10,7 +10,6 @@ use Mouse;
 use JSON;
 use URI;
 
-has isPP                     => ( is => 'rw' );
 has speChars                 => ( is => 'rw' );
 has skinRules                => ( is => 'rw' );
 has stayConnected            => ( is => 'rw', default => sub { 0 } );
@@ -21,34 +20,22 @@ has passwordResetUrl => (
     is      => 'ro',
     lazy    => 1,
     default => sub {
-        my $self = $_[0];
-        my $p    = $self->conf->{portal};
-        $p =~ s#/*$##;
-        return $self->conf->{mailUrl} ? $self->conf->{mailUrl} : "$p/resetpwd";
+        $_[0]->conf->{mailUrl} || $_[0]->buildUrl("resetpwd");
     }
 );
 has certificateResetUrl => (
     is      => 'ro',
     lazy    => 1,
     default => sub {
-        my $self = $_[0];
-        my $p    = $self->conf->{portal};
-        $p =~ s#/*$##;
-        return $self->conf->{certificateResetByMailURL}
-          ? $self->conf->{certificateResetByMailURL}
-          : "$p/certificateReset";
+        $_[0]->conf->{certificateResetByMailURL}
+          || $_[0]->buildUrl("certificateReset");
     }
 );
 has registerUrl => (
     is      => 'ro',
     lazy    => 1,
     default => sub {
-        my $self = $_[0];
-        my $p    = $self->conf->{portal};
-        $p =~ s#/*$##;
-        return $self->conf->{registerUrl}
-          ? $self->conf->{registerUrl}
-          : "$p/register";
+        $_[0]->conf->{registerUrl} || $_[0]->buildUrl("register");
     }
 );
 
@@ -110,13 +97,19 @@ sub displayInit {
       : map { ( s/\s+/<space>/r => 1 ) } split '',
       $self->conf->{passwordPolicySpecialChar};
     $self->speChars( join q{ }, sort keys %speChars );
+}
 
-    $self->isPP( $self->speChars
-          || $self->conf->{passwordPolicyMinSize}
-          || $self->conf->{passwordPolicyMinLower}
-          || $self->conf->{passwordPolicyMinUpper}
-          || $self->conf->{passwordPolicyMinDigit}
-          || $self->conf->{passwordPolicyMinSpeChar} );
+sub isPP {
+    my $self = shift;
+    my $ppRules = $self->getPpolicyRules();
+    foreach my $rule ( @$ppRules )
+    {
+        if( $rule->{'condition'} )
+        {
+            return 1;
+        }
+    }
+    return 0;
 }
 
 # Call portal process and set template parameters
@@ -156,18 +149,19 @@ sub display {
         $self->logger->debug('Display: confirm detected');
         $skinfile       = 'confirm';
         %templateParams = (
-            AUTH_ERROR      => $req->error,
-            AUTH_ERROR_TYPE => $req->error_type,
-            AUTH_ERROR_ROLE => $req->error_role,
-            AUTH_URL        => $req->{data}->{_url},
-            MSG             => $req->info,
-            HIDDEN_INPUTS   => $self->buildHiddenForm($req),
-            ACTIVE_TIMER    => $req->data->{activeTimer},
-            FORM_ACTION     => $req->data->{confirmFormAction} || "#",
-            FORM_METHOD     => $self->conf->{confirmFormMethod},
-            CHOICE_PARAM    => $self->conf->{authChoiceParam},
-            CHOICE_VALUE    => $req->data->{_authChoice},
-            CHECK_LOGINS    => $self->conf->{portalCheckLogins}
+            AUTH_ERROR                      => $req->error,
+            AUTH_ERROR_TYPE                 => $req->error_type,
+            AUTH_ERROR_ROLE                 => $req->error_role,
+            ( 'AUTH_ERROR_' . $req->error ) => 1,
+            AUTH_URL                        => $req->{data}->{_url},
+            MSG                             => $req->info,
+            HIDDEN_INPUTS                   => $self->buildHiddenForm($req),
+            ACTIVE_TIMER                    => $req->data->{activeTimer},
+            FORM_ACTION  => $req->data->{confirmFormAction} || "#",
+            FORM_METHOD  => $self->conf->{confirmFormMethod},
+            CHOICE_PARAM => $self->conf->{authChoiceParam},
+            CHOICE_VALUE => $req->data->{_authChoice},
+            CHECK_LOGINS => $self->conf->{portalCheckLogins}
               && $req->data->{login},
             ASK_LOGINS        => $req->param('checkLogins')   || 0,
             ASK_STAYCONNECTED => $req->param('stayconnected') || 0,
@@ -186,16 +180,17 @@ sub display {
         $self->logger->debug('Display: IDP choice detected');
         $skinfile       = 'idpchoice';
         %templateParams = (
-            AUTH_ERROR      => $req->error,
-            AUTH_ERROR_TYPE => $req->error_type,
-            AUTH_ERROR_ROLE => $req->error_role,
-            AUTH_URL        => $req->{data}->{_url},
-            HIDDEN_INPUTS   => $self->buildHiddenForm($req),
-            ACTIVE_TIMER    => $req->data->{activeTimer},
-            FORM_METHOD     => $self->conf->{confirmFormMethod},
-            CHOICE_PARAM    => $self->conf->{authChoiceParam},
-            CHOICE_VALUE    => $req->data->{_authChoice},
-            CHECK_LOGINS    => $self->conf->{portalCheckLogins}
+            AUTH_ERROR                      => $req->error,
+            AUTH_ERROR_TYPE                 => $req->error_type,
+            AUTH_ERROR_ROLE                 => $req->error_role,
+            ( 'AUTH_ERROR_' . $req->error ) => 1,
+            AUTH_URL                        => $req->{data}->{_url},
+            HIDDEN_INPUTS                   => $self->buildHiddenForm($req),
+            ACTIVE_TIMER                    => $req->data->{activeTimer},
+            FORM_METHOD                     => $self->conf->{confirmFormMethod},
+            CHOICE_PARAM                    => $self->conf->{authChoiceParam},
+            CHOICE_VALUE                    => $req->data->{_authChoice},
+            CHECK_LOGINS                    => $self->conf->{portalCheckLogins}
               && $req->data->{login},
             ASK_LOGINS        => $req->param('checkLogins')   || 0,
             ASK_STAYCONNECTED => $req->param('stayconnected') || 0,
@@ -219,11 +214,12 @@ sub display {
           for keys %{ $req->{portalHiddenFormValues} // {} };
         $skinfile       = 'info';
         %templateParams = (
-            AUTH_ERROR      => $self->error,
-            AUTH_ERROR_TYPE => $req->error_type,
-            AUTH_ERROR_ROLE => $req->error_role,
-            MSG             => $info,
-            URL           => $req->{urldc} || $self->conf->{portal},  # Fix 2158
+            AUTH_ERROR                      => $req->error,
+            AUTH_ERROR_TYPE                 => $req->error_type,
+            AUTH_ERROR_ROLE                 => $req->error_role,
+            ( 'AUTH_ERROR_' . $req->error ) => 1,
+            MSG                             => $info,
+            URL           => $req->{urldc} || $req->portal,  # Fix 2158
             HIDDEN_INPUTS => $self->buildOutgoingHiddenForm( $req, $method ),
             ACTIVE_TIMER  => $req->data->{activeTimer},
             CHOICE_PARAM  => $self->conf->{authChoiceParam},
@@ -246,16 +242,18 @@ sub display {
         or $req->{error} == PE_OPENID_BADID )
     {
         $skinfile = 'openid';
-        my $p = $self->conf->{portal} . $self->conf->{issuerDBOpenIDPath};
-        $p =~ s#(?<!:)/?\^?/#/#g;
+
+        my $openid_path =
+          $self->conf->{issuerDBOpenIDPath} =~ s/^.*?(\w+).*?$/$1/r;
         my $id = $req->{sessionInfo}
           ->{ $self->conf->{openIdAttr} || $self->conf->{whatToTrace} };
         %templateParams = (
-            AUTH_ERROR      => $self->error,
-            AUTH_ERROR_TYPE => $req->error_type,
-            AUTH_ERROR_ROLE => $req->error_role,
-            PROVIDERURI     => $p,
-            MSG             => $req->info(),
+            AUTH_ERROR                      => $self->error,
+            AUTH_ERROR_TYPE                 => $req->error_type,
+            AUTH_ERROR_ROLE                 => $req->error_role,
+            ( 'AUTH_ERROR_' . $req->error ) => 1,
+            PROVIDERURI => $self->buildUrl( $req->portal, $openid_path, "" ),
+            MSG         => $req->info(),
             (
                 $req->data->{customScript}
                 ? ( CUSTOM_SCRIPT => $req->data->{customScript} )
@@ -283,6 +281,22 @@ sub display {
         );
     }
 
+    # 2.1 Wait for 2FA
+    elsif ( $req->{error} == PE_2FWAIT ) {
+        my $method = $req->data->{redirectFormMethod} || 'get';
+        $skinfile       = "2fwait";
+        %templateParams = (
+            URL           => $req->{urldc},
+            HIDDEN_INPUTS => $self->buildOutgoingHiddenForm( $req, $method ),
+            FORM_METHOD   => $method,
+            (
+                $req->data->{customScript}
+                ? ( CUSTOM_SCRIPT => $req->data->{customScript} )
+                : ()
+            ),
+        );
+    }
+
     # 2.2 Case : display menu (with error or not)
     elsif ( $req->error == PE_OK ) {
         $skinfile = 'menu';
@@ -291,32 +305,14 @@ sub display {
         %templateParams = (
             AUTH_USER => $req->{sessionInfo}->{ $self->conf->{portalUserAttr} },
             NEWWINDOW => $self->conf->{portalOpenLinkInNewWindow},
-            LOGOUT_URL              => $self->conf->{portal} . "?logout=1",
+            LOGOUT_URL              => $self->buildUrl( $req->portal, { logout => 1 } ),
             APPSLIST_ORDER          => $req->{sessionInfo}->{'_appsListOrder'},
             PING                    => $self->conf->{portalPingInterval},
             DONT_STORE_PASSWORD     => $self->conf->{browsersDontStorePassword},
             HIDE_OLDPASSWORD        => 0,
-            PPOLICY_NOPOLICY        => !$self->isPP(),
             ENABLE_PASSWORD_DISPLAY =>
               $self->conf->{portalEnablePasswordDisplay},
-            ENABLE_CHECKHIBP => $self->conf->{checkHIBP},
-            DISPLAY_PPOLICY  => $self->conf->{portalDisplayPasswordPolicy},
-            PPOLICY_MINSIZE  => $self->conf->{passwordPolicyMinSize},
-            PPOLICY_MINLOWER => $self->conf->{passwordPolicyMinLower},
-            PPOLICY_MINUPPER => $self->conf->{passwordPolicyMinUpper},
-            PPOLICY_MINDIGIT => $self->conf->{passwordPolicyMinDigit},
-            (
-                $self->conf->{passwordPolicyMinSpeChar}
-                  && $self->conf->{passwordPolicySpecialChar}
-                ? ( PPOLICY_MINSPECHAR =>
-                      $self->conf->{passwordPolicyMinSpeChar} )
-                : ()
-            ),
-            (
-                $self->conf->{passwordPolicyMinSpeChar} || $self->speChars()
-                ? ( PPOLICY_ALLOWEDSPECHAR => $self->speChars() )
-                : ()
-            ),
+            %{ $self->getPasswordPolicyTemplateVars },
             (
                 $self->requireOldPwd->( $req, $req->userData )
                 ? ( REQUIRE_OLDPASSWORD => 1 )
@@ -341,7 +337,7 @@ sub display {
             PORTALBUTTON => 1,
             BUTTON       => 'upgradeSession',
             CONFIRMKEY   => $self->stamp,
-            PORTAL       => $self->conf->{portal},
+            PORTAL       => $req->portal,
             URL          => $req->data->{_url},
             (
                 $req->data->{customScript}
@@ -358,7 +354,7 @@ sub display {
             FORMACTION   => '/renewsession',
             MSG          => 'askToRenew',
             CONFIRMKEY   => $self->stamp,
-            PORTAL       => $self->conf->{portal},
+            PORTAL       => $req->portal,
             PORTALBUTTON => 1,
             BUTTON       => 'renewSession',
             URL          => $req->data->{_url},
@@ -378,7 +374,7 @@ sub display {
             MSG        => 'PE87',
             CONFIRMKEY => $self->stamp,
             BUTTON     => 'renewSession',
-            PORTAL     => $self->conf->{portal},
+            PORTAL     => $req->portal,
             URL        => $req->data->{_url},
             (
                 $req->data->{customScript}
@@ -406,10 +402,11 @@ sub display {
     {
         $skinfile       = 'error';
         %templateParams = (
-            AUTH_ERROR      => $req->error,
-            AUTH_ERROR_TYPE => $req->error_type,
-            AUTH_ERROR_ROLE => $req->error_role,
-            LOCKTIME        => $req->lockTime(),
+            AUTH_ERROR                      => $req->error,
+            AUTH_ERROR_TYPE                 => $req->error_type,
+            AUTH_ERROR_ROLE                 => $req->error_role,
+            ( 'AUTH_ERROR_' . $req->error ) => 1,
+            LOCKTIME                        => $req->lockTime(),
             (
                 $req->data->{customScript}
                 ? ( CUSTOM_SCRIPT => $req->data->{customScript} )
@@ -423,11 +420,12 @@ sub display {
         $skinfile = 'login';
         my $login = $req->user;
         %templateParams = (
-            AUTH_ERROR            => $req->error,
-            AUTH_ERROR_TYPE       => $req->error_type,
-            AUTH_ERROR_ROLE       => $req->error_role,
-            AUTH_URL              => $req->{data}->{_url},
-            LOGIN                 => $login,
+            AUTH_ERROR                      => $req->error,
+            AUTH_ERROR_TYPE                 => $req->error_type,
+            AUTH_ERROR_ROLE                 => $req->error_role,
+            ( 'AUTH_ERROR_' . $req->error ) => 1,
+            AUTH_URL                        => $req->{data}->{_url},
+            LOGIN                           => $login,
             DONT_STORE_PASSWORD   => $self->conf->{browsersDontStorePassword},
             CHECK_LOGINS          => $self->conf->{portalCheckLogins},
             ASK_LOGINS            => $req->param('checkLogins')   || 0,
@@ -444,10 +442,11 @@ sub display {
               || $self->conf->{proxyAuthServiceImpersonation},
             ENABLE_PASSWORD_DISPLAY =>
               $self->conf->{portalEnablePasswordDisplay},
-            ENABLE_CHECKHIBP => $self->conf->{checkHIBP},
-            (
-                $self->stayConnected->( $req, $req->sessionInfo )
-                ? ( STAYCONNECTED => 1 )
+            %{ $self->getPasswordPolicyTemplateVars },
+            ( (
+                         $self->conf->{trustedBrowserRule}
+                      or $self->stayConnected->( $req, $req->sessionInfo )
+                ) ? ( STAYCONNECTED => 1 )
                 : ()
             ),
             (
@@ -470,15 +469,15 @@ sub display {
         # External links
         if ( $self->conf->{portalDisplayResetPassword} ) {
             $templateParams{"MAIL_URL_EXTERNAL"} =
-              $self->_isExternalUrl( $self->passwordResetUrl );
+              $self->_isExternalUrl( $req, $self->passwordResetUrl );
         }
         if ( $self->conf->{portalDisplayRegister} ) {
             $templateParams{"REGISTER_URL_EXTERNAL"} =
-              $self->_isExternalUrl( $self->registerUrl );
+              $self->_isExternalUrl( $req, $self->registerUrl );
         }
         if ( $self->conf->{portalDisplayCertificateResetByMail} ) {
             $templateParams{MAILCERTIF_URL_EXTERNAL} =
-              $self->_isExternalUrl( $self->certificateResetUrl );
+              $self->_isExternalUrl( $req, $self->certificateResetUrl );
         }
 
         # Display captcha if it's enabled
@@ -528,31 +527,14 @@ sub display {
                 CHOICE_PARAM          => $self->conf->{authChoiceParam},
                 CHOICE_VALUE          => $req->data->{_authChoice},
                 OLDPASSWORD           => $self->checkXSSAttack( 'oldpassword',
-                    $req->data->{oldpassword} ) ? ''
+                    $req->data->{oldpassword} )
+                ? ''
                 : $req->data->{oldpassword},
                 HIDE_OLDPASSWORD    => $self->conf->{hideOldPassword},
                 DONT_STORE_PASSWORD => $self->conf->{browsersDontStorePassword},
-                PPOLICY_NOPOLICY    => !$self->isPP(),
                 ENABLE_PASSWORD_DISPLAY =>
                   $self->conf->{portalEnablePasswordDisplay},
-                ENABLE_CHECKHIBP => $self->conf->{checkHIBP},
-                DISPLAY_PPOLICY  => $self->conf->{portalDisplayPasswordPolicy},
-                PPOLICY_MINSIZE  => $self->conf->{passwordPolicyMinSize},
-                PPOLICY_MINLOWER => $self->conf->{passwordPolicyMinLower},
-                PPOLICY_MINUPPER => $self->conf->{passwordPolicyMinUpper},
-                PPOLICY_MINDIGIT => $self->conf->{passwordPolicyMinDigit},
-                (
-                    $self->conf->{passwordPolicyMinSpeChar}
-                      && $self->conf->{passwordPolicySpecialChar}
-                    ? ( PPOLICY_MINSPECHAR =>
-                          $self->conf->{passwordPolicyMinSpeChar} )
-                    : ()
-                ),
-                (
-                    $self->conf->{passwordPolicyMinSpeChar} || $self->speChars()
-                    ? ( PPOLICY_ALLOWEDSPECHAR => $self->speChars() )
-                    : ()
-                ),
+                %{ $self->getPasswordPolicyTemplateVars },
             );
         }
 
@@ -608,7 +590,16 @@ sub display {
                     DISPLAY_FINDUSER     => scalar @$fields,
                     MANDATORY            => $mandatory,
                     FIELDS               => $fields,
-                    SPOOFID              => $slogin
+                    SPOOFID              => $slogin,
+
+                    # Hidden inputs and custom scripts may have been modified
+                    # by _buildAuthLoop so we need to update them
+                    HIDDEN_INPUTS => $self->buildHiddenForm($req),
+                    (
+                        $req->data->{customScript}
+                        ? ( CUSTOM_SCRIPT => $req->data->{customScript} )
+                        : ()
+                    ),
                 );
             }
 
@@ -638,7 +629,7 @@ sub display {
                     : "",
                     AUTH_LOOP  => [],
                     PORTAL_URL =>
-                      ( $displayType eq "logo" ? $self->conf->{portal} : 0 ),
+                      ( $displayType eq "logo" ? $req->portal : 0 ),
                     MSG       => $req->info(),
                     MANDATORY => $mandatory,
                     FIELDS    => $fields,
@@ -684,30 +675,18 @@ sub staticFile {
 
 sub buildOutgoingHiddenForm {
     my ( $self, $req, $method ) = @_;
-    my @keys = keys %{ $req->{portalHiddenFormValues} };
+    my @extra_keys;
 
     if ( lc $method eq 'get' ) {
-        my $uri          = URI->new( $req->{urldc} );
-        my %query_params = $uri->query_form;
-
-        # Redirection URL contains query string. Before displaying a form,
-        # we must set the query string parameters as form fields so they can
-        # be preserved #2085
-        if (%query_params) {
-            $self->logger->debug(
-"urldc contains query parameters, setting them as hidden form values"
-            );
-            foreach ( keys %query_params ) {
-                $self->setHiddenFormValue( $req, $_, $query_params{$_}, "", 0 );
-            }
-        }
+        my $uri = URI->new( $req->{urldc} );
+        @extra_keys = $uri->query_form;
     }
 
-    return $self->buildHiddenForm($req);
+    return $self->buildHiddenForm( $req, @extra_keys );
 }
 
 sub buildHiddenForm {
-    my ( $self, $req ) = @_;
+    my ( $self, $req, @extra_keys ) = @_;
     my @keys = keys %{ $req->{portalHiddenFormValues} };
     my $val  = '';
 
@@ -719,13 +698,20 @@ sub buildHiddenForm {
             && $self->checkXSSAttack( $_, $req->{portalHiddenFormValues}->{$_} )
           );
 
-        # Build hidden input HTML code
-        # 'id' is removed to avoid warning with Choice
-        #$val .= qq{<input type="hidden" name="$_" id="$_" value="}
         $val .= qq{<input type="hidden" name="$_" value="}
           . $req->{portalHiddenFormValues}->{$_} . '" />';
     }
 
+    while ( my ( $key, $value ) = splice( @extra_keys, 0, 2 ) ) {
+
+        # Check XSS attacks
+        next
+          if (!$req->data->{safeHiddenFormValues}->{$key}
+            && $self->checkXSSAttack( $key, $value ) );
+
+        $val .= qq{<input type="hidden" name="$key" value="$value" />};
+    }
+
     return $val;
 }
 
@@ -759,18 +745,18 @@ sub getSkin {
     my $skinParam = $req->param('skin');
     if ( defined $skinParam ) {
         if ( $skinParam =~ /^[\w\-]+$/ ) {
-            if ( -d $self->conf->{templateDir} . '/' . $skinParam ) {
+            if ( -d $self->getSkinTplDir( $req, $skinParam ) ) {
                 $skin = $skinParam;
                 $self->logger->debug(
                     "Skin $skin selected from GET/POST parameter");
             }
             else {
-                $self->userLogger->error(
-                    "User tries to access to unexistent skin dir $skinParam");
+                $self->logger->warn(
+                    "User tries to access to nonexistent skin dir $skinParam");
             }
         }
         else {
-            $self->userLogger->error("Strange skin parameter: $skinParam");
+            $self->logger->warn("Invalid skin name $skinParam");
         }
     }
 
@@ -820,7 +806,14 @@ sub mkSessionArray {
                         ip     => $session->{ipAddr},
                         values => [
                             map {
-                                { v => $session->{$_}, k => $_, "k_$_" => 1 }
+                                # Modifying key to remove ordering prefix
+                                my $modifiedKey = $_;
+                                $modifiedKey =~ s/(\d+_)?//;
+                                {
+                                    v => $session->{$modifiedKey},
+                                    k => $modifiedKey,
+                                    "k_$modifiedKey" => 1
+                                }
                             } @fields
                         ],
                         error        => $session->{error},
@@ -896,8 +889,64 @@ sub mkOidcConsent {
 }
 
 sub _isExternalUrl {
-    my ( $self, $url ) = @_;
-    return ( index( $url, $self->conf->{portal} ) < 0 );
+    my ( $self, $req, $url ) = @_;
+    return ( index( $url, $req->portal ) < 0 );
+}
+
+sub getPasswordPolicyTemplateVars {
+    my ( $self, $req ) = @_;
+    return {
+        PPOLICY_RULES    => $self->getPpolicyRules,
+        PPOLICY_NOPOLICY => !$self->isPP(),
+        ENABLE_CHECKHIBP => $self->conf->{checkHIBP},
+        DISPLAY_PPOLICY  => $self->conf->{portalDisplayPasswordPolicy},
+        PPOLICY_MINSIZE  => $self->conf->{passwordPolicyMinSize},
+        PPOLICY_MINLOWER => $self->conf->{passwordPolicyMinLower},
+        PPOLICY_MINUPPER => $self->conf->{passwordPolicyMinUpper},
+        PPOLICY_MINDIGIT => $self->conf->{passwordPolicyMinDigit},
+        (
+            $self->conf->{passwordPolicyMinSpeChar}
+            ? ( PPOLICY_MINSPECHAR => $self->conf->{passwordPolicyMinSpeChar} )
+            : ()
+        ),
+        (
+            $self->conf->{passwordPolicyMinSpeChar} && $self->speChars()
+            ? (
+                PPOLICY_ALLOWEDSPECHAR      => $self->speChars(),
+                PPOLICY_ALLOWEDSPECHAR_JSON =>
+                  to_json( $self->speChars(), { allow_nonref => 1 } ),
+              )
+            : ()
+        ),
+    };
+}
+
+sub getPpolicyRules {
+    my ($self) = @_;
+    my $result = [];
+
+    # Avoid useless computation if ppolicy is not displayed
+    return $result if !$self->conf->{portalDisplayPasswordPolicy};
+
+    while ( my ( $id, $policy ) = each %{ $self->_ppRules } ) {
+        push @$result, { %$policy, id => $id };
+    }
+
+    # Sort ppolicy items by "order" numerically and then by "id" lexically
+    $result = [ sort { $a->{order} <=> $b->{order} || $a->{id} cmp $b->{id} } @$result ];
+
+    # Format data attributes for HTML::Template loop
+    # data => { foo => "bar" } becomes
+    # data => [ { key => "foo", value => "bar" } ]
+    for my $result (@$result) {
+        if ( ref( $result->{data} ) eq "HASH" ) {
+            $result->{data} = [
+                map { { key => $_, value => $result->{data}->{$_} } }
+                  keys %{ $result->{data} }
+            ];
+        }
+    }
+    return $result;
 }
 
 1;
